@@ -13,10 +13,12 @@ Covering areas:
 import argparse
 import collections
 import cProfile as profile
+import copy
 import difflib
 import fnmatch
 import functools
 import gettext
+import glob
 import hashlib
 import json
 import locale
@@ -26,6 +28,7 @@ import multiprocessing
 import operator
 import os
 import os.path as osp
+import types
 from os.path import abspath, basename, dirname, expanduser, exists, isfile, join, splitext
 # from pprint import pprint, pformat
 import platform
@@ -42,6 +45,7 @@ import threading
 import traceback
 from types import SimpleNamespace
 import uuid
+import warnings
 
 
 #
@@ -814,35 +818,39 @@ class RerunLock:
         filename = f'lock_{name}.json' if name else 'lock_{}.json'.format(next(tempfile._get_candidate_names()))
         self.lockFile = osp.join(folder, filename) if folder else join(get_platform_tmp_dir(), filename)
         self.logger = logger
-        common_sigs = [
-            signal.SIGABRT,
-            signal.SIGFPE,
-            signal.SIGILL,
-            signal.SIGINT,
-            signal.SIGSEGV,
-            signal.SIGTERM,
-        ]
-        plat_sigs = [
-            signal.SIGBREAK,
-            # CAUTION
-            # - CTRL_C_EVENT, CTRL_BREAK_EVENT not working on Windows
-            # signal.CTRL_C_EVENT,
-            # signal.CTRL_BREAK_EVENT,
-        ] if platform.system() == 'Windows' else [
-            # CAUTION:
-            # - SIGCHLD as an alias is safe to ignore
-            # - SIGKILL must be handled by os.kill()
-            signal.SIGALRM,
-            signal.SIGBUS,
-            # signal.SIGCHLD,
-            signal.SIGCONT,
-            signal.SIGHUP,
-            # signal.SIGKILL,
-            signal.SIGPIPE,
-        ]
-        for sig in common_sigs + plat_sigs:
-            # breakpoint()
-            signal.signal(sig, self.handle_signal)
+        # CAUTION:
+        # - windows grpc server crashes with signals:
+        #   - ValueError: signal only works in main thread of the main interpreter
+        # - signals are disabled for windows
+        if threading.current_thread() is threading.main_thread():
+            common_sigs = [
+                signal.SIGABRT,
+                signal.SIGFPE,
+                signal.SIGILL,
+                signal.SIGINT,
+                signal.SIGSEGV,
+                signal.SIGTERM,
+            ]
+            plat_sigs = [
+                signal.SIGBREAK,
+                # CAUTION
+                # - CTRL_C_EVENT, CTRL_BREAK_EVENT not working on Windows
+                # signal.CTRL_C_EVENT,
+                # signal.CTRL_BREAK_EVENT,
+            ] if platform.system() == 'Windows' else [
+                # CAUTION:
+                # - SIGCHLD as an alias is safe to ignore
+                # - SIGKILL must be handled by os.kill()
+                signal.SIGALRM,
+                signal.SIGBUS,
+                # signal.SIGCHLD,
+                signal.SIGCONT,
+                signal.SIGHUP,
+                # signal.SIGKILL,
+                signal.SIGPIPE,
+            ]
+            for sig in common_sigs + plat_sigs:
+                signal.signal(sig, self.handle_signal)
 
     def lock(self):
         if not self.is_locked():
@@ -1233,6 +1241,52 @@ def find_first_line_in_range(lines, keyword, linerange=(0,), algo='startswith'):
     return lineno_in_between + linerange[0] if lineno_in_between is not None else None
 
 
+def substitute_lines_between_cues(inserts, iolines, startcue, endcue, startlineno=0, removecues=False, withindent=True, useappend=False, skipdups=False):
+    """
+    - lazy-create list if input lines is a string (a single line)
+    - smart-indent lines according to tag indentation
+    - optimize with search range slicing
+    - do not add lineends, treat everything as generic strings
+    - returns indices of inserted in resulted lines
+    """
+    inserts = [inserts] if isinstance(inserts, str) else inserts
+    selected_lines = iolines[startlineno:] if startlineno > 0 else iolines
+    # find range
+    startln, endln = None, None
+    # always use startswith, because when we leave a cue, we want it to appear first and foremost
+    startln = next((ln for ln, line in enumerate(selected_lines) if line.strip().startswith(startcue)), None)
+    if startln is None:
+        return startln, endln
+    endln = next((ln for ln, line in enumerate(selected_lines[startln:]) if line.strip().startswith(endcue)), None)
+    if endln is None:
+        return startlineno+startln, None
+    if removecues:
+        startln -= 1
+        endln += 2
+    # back to all lines with offset applied
+    startln += startlineno
+    endln += startln
+    if withindent:
+        ref_startln = startln+1 if removecues else startln
+        indent = iolines[ref_startln].find(startcue)
+        indent_by_spaces = 0
+        for idt in range(indent):
+            indent_by_spaces += 4 if iolines[ref_startln][idt] == '\t' else 1
+        inserts = ['{}{}'.format(' ' * indent_by_spaces, line) for line in inserts]
+    # append to current content between cues or not
+    lines_to_insert = iolines[startln+1: endln] + inserts if useappend else inserts
+    if skipdups:
+        lines_to_insert = list(dict.fromkeys(lines_to_insert))
+    # remove lines in b/w
+    has_lines_between_keywords = endln - startln > 1
+    if has_lines_between_keywords:
+        del iolines[startln+1: endln]
+    iolines[startln+1: startln+1] = lines_to_insert
+    insert_start = startln+1
+    rg_inserted = [insert_start, insert_start+len(lines_to_insert)-1]
+    return rg_inserted
+
+
 def substitute_lines_between_keywords(lines, file, opkey, edkey, startlineno=0, withindent=True, useappend=False, skipdups=False):
     """
     - lazy-append line-ends to input lines
@@ -1241,6 +1295,7 @@ def substitute_lines_between_keywords(lines, file, opkey, edkey, startlineno=0,
     - optimize with search range slicing
     - returns original indices
     """
+    warnings.warn(deprecate_log('substitute_lines_between_cues()'), DeprecationWarning, stacklevel=2)
     lines = [lines] if isinstance(lines, str) else lines
     lines = [line if line.endswith('\n') else f'{line}\n' for line in lines]
     with open(file) as fp:
@@ -1248,34 +1303,34 @@ def substitute_lines_between_keywords(lines, file, opkey, edkey, startlineno=0,
     selected_lines = all_lines[startlineno:] if startlineno > 0 else all_lines
     # find range
     rg_insert = [None, None]
-    rg_insert[0] = next((li for li, line in enumerate(selected_lines) if line.strip().startswith(opkey)), None)
-    if rg_insert[0] is None:
+    startln = next((li for li, line in enumerate(selected_lines) if line.strip().startswith(opkey)), None)
+    if startln is None:
         return rg_insert
-    rg_insert[1] = next((li for li, line in enumerate(selected_lines[rg_insert[0]:]) if line.strip().startswith(edkey)), None)
-    if rg_insert[1] is None:
-        return startlineno+rg_insert[0], None
+    endln = next((li for li, line in enumerate(selected_lines[startln:]) if line.strip().startswith(edkey)), None)
+    if endln is None:
+        return startlineno+startln, None
     # back to all lines with offset applied
-    rg_insert[0] += startlineno
-    rg_insert[1] += rg_insert[0]
+    startln += startlineno
+    endln += startln
     if withindent:
-        indent = all_lines[rg_insert[0]].find(opkey)
+        indent = all_lines[startln].find(opkey)
         indent_by_spaces = 0
         for idt in range(indent):
-            indent_by_spaces += 4 if all_lines[rg_insert[0]][idt] == '\t' else 1
+            indent_by_spaces += 4 if all_lines[startln][idt] == '\t' else 1
         assert indent_by_spaces >= 0
         lines = ['{}{}'.format(' '*indent_by_spaces, line) for line in lines]
     # remove duplicates
-    lines_to_insert = all_lines[rg_insert[0]+1: rg_insert[1]] + lines if useappend else lines
+    lines_to_insert = all_lines[startln+1: endln] + lines if useappend else lines
     if skipdups:
         lines_to_insert = list(dict.fromkeys(lines_to_insert))
     # remove lines in b/w
-    has_lines_between_keywords = rg_insert[1] - rg_insert[0] > 1
+    has_lines_between_keywords = endln - startln > 1
     if has_lines_between_keywords:
-        del all_lines[rg_insert[0]+1: rg_insert[1]]
-    all_lines[rg_insert[0]+1: rg_insert[0]+1] = lines_to_insert
+        del all_lines[startln+1: endln]
+    all_lines[startln+1: startln+1] = lines_to_insert
     with open(file, 'w') as fp:
         fp.writelines(all_lines)
-    rg_inserted = [rg_insert[0], rg_insert[0]+len(lines_to_insert)]
+    rg_inserted = [startln, startln+len(lines_to_insert)]
     return rg_inserted
 
 
@@ -1372,7 +1427,7 @@ def duplicate_dir(srcdir, dstdir):
     _dup_dir_posix(srcdir, dstdir)
 
 
-def compare_textfiles(file1, file2, showdiff=False, contextonly=True, logger=None):
+def compare_textfiles(file1, file2, showdiff=False, contextonly=True, ignoredlinenos=None, logger=None):
     with open(file1) as fp1, open(file2) as fp2:
         lines1 = fp1.readlines()
         lines2 = fp2.readlines()
@@ -1381,6 +1436,9 @@ def compare_textfiles(file1, file2, showdiff=False, contextonly=True, logger=Non
             diff = diff_func(lines1, lines2)
             lazy_logging(f'***\n{file1} vs.\n{file2}\n***')
             lazy_logging(''.join(diff), logger)
+    if ignoredlinenos:
+        lines1 = [line for ln, line in enumerate(lines1) if ln not in ignoredlinenos]
+        lines2 = [line for ln, line in enumerate(lines2) if ln not in ignoredlinenos]
     return lines1 == lines2
 
 
@@ -1391,14 +1449,15 @@ def lazy_logging(msg, logger=None):
         print(msg)
 
 
-def copy_file(src, dst, isdstdir=False):
+def copy_file(src, dst, isdstdir=False, keepmeta=False):
     if isdstdir:
         par_dir = dst
     else:
         par_dir = osp.split(dst)[0]
     os.makedirs(par_dir, exist_ok=True)
+    copyfunc = shutil.copy2 if keepmeta else shutil.copy
     try:
-        shutil.copy(src, dst)
+        copyfunc(src, dst)
     except shutil.SameFileError:
         _logger.warning(f'source and destination are identical. will SKIP: {osp.abspath(src)} -> {osp.abspath(dst)}.')
 
@@ -1456,8 +1515,7 @@ def pack_obj(obj, topic=None, envelope=('<KK-ENV>', '</KK-ENV>'), classes=()):
     """
     if not topic:
         topic = type(obj).__name__
-    from types import SimpleNamespace
-    msg = SimpleNamespace(payload=obj, topic=topic)
+    msg = types.SimpleNamespace(payload=obj, topic=topic)
     if classes:
         class CustomJsonEncoder(json.JSONEncoder):
             def default(self, o):
@@ -1484,6 +1542,21 @@ def get_parent_dirs(file, subs=(), depth=1):
     return script_dir, root, *[osp.join(root, sub) for sub in subs]
 
 
+def get_ancestor_dirs(file, depth=1):
+    """
+    given structure: X > Y > Z > file,
+    return folder sequence: Z, Y, X
+    """
+    par_dir = osp.abspath(osp.dirname(file))
+    if depth < 2:
+        return par_dir
+    dirs = [par_dir]
+    # focus on pardir(i.e., depth 1), then backtrace 1 at a time from depth-1 to depth
+    for dp in range(depth-1):
+        dirs.append(osp.abspath(osp.join(par_dir, osp.normpath('../'*(dp+1)))))
+    return dirs
+
+
 def get_child_dirs(root, subs=()):
     return (osp.join(root, sub) for sub in subs)
 
@@ -1527,17 +1600,130 @@ def flatten_nested_lists(mylist):
 def show_results(succeeded, detail, advice, dryrun=False):
     banner = '** DRYRUN **' if dryrun else '*** SUCCEEDED ***' if succeeded else '* FAILED *'
     advice_title = '' if not advice else 'Next:' if succeeded else 'Advice:'
-    report = f"""
+    detail_block = f'\n{detail}\n' if detail and advice else f'\n{detail}' if detail else ''
+    if detail and advice:
+        report = f"""
 {banner}
-
-{detail}
-
+{detail_block}
 {advice_title}
-{advice}
-"""
+{advice}"""
+    elif detail:
+        report = f"""
+{banner}
+{detail_block}"""
+    elif advice:
+        report = f"""
+{banner}
+{advice_title}
+{advice}"""
+    else:
+        report = f"""
+{banner}"""
     print(report)
 
 
+def init_repo(srcfile, appdepth=2, repodepth=3, organization='mycompany', verbose=False, uselocale=False):
+    """
+    assuming a project has a folder structure, create structure and facilities around it
+    - structure example: app > subs (src, test, temp, locale, ...), where app is app root
+    - structure example: repo > app > subs (src, test, temp, locale, ...), where repo is app-suite root
+    - by default, assume srcfile is under repo > app > src
+    - set flag uselocale to use gettext localization, by using _T() function around non-fstrings
+    - set verbose to all show log levels in console
+    - set inter-app sharable tmp folder to platform_cache > organization > app
+    """
+    assert appdepth <= repodepth
+    common = types.SimpleNamespace()
+    common.ancestorDirs = get_ancestor_dirs(srcfile, depth=repodepth)
+    # include repo so that import can use repo-based dot syntax
+    lazy_extend_sys_path([repo_root := common.ancestorDirs[repodepth - 1]])
+    # just have fixed initial folders to meet most needs in core and tests
+    common.locDir, common.srcDir, common.tmpDir, common.testDir = get_child_dirs(app_root := common.ancestorDirs[appdepth - 1], subs=('locale', 'src', 'temp', 'test'))
+    common.pubTmpDir = osp.join(get_platform_tmp_dir(), organization, osp.basename(app_root))
+    common.stem = osp.splitext(osp.basename(srcfile))[0]
+    common.logger = build_default_logger(common.tmpDir, name=common.stem, verbose=verbose)
+    if uselocale:
+        common.translator = init_translator(common.locDir)
+    return common
+
+
+def backup_file(file, dstdir=None, suffix='.1', keepmeta=True):
+    """
+    save numeric backup in dstdir or same dir
+    - preserve metadata
+    - always overwrite non-numeric backup
+    """
+    bak_dir = dstdir if dstdir else osp.dirname(file)
+    bak = osp.join(bak_dir, osp.basename(file)+suffix)
+    num = suffix[1:]
+    if not num.isnumeric():
+        copy_file(file, bak, keepmeta=keepmeta)
+        return
+    while osp.isfile(bak):
+        stem = osp.splitext(bak)[0]
+        num = int(osp.splitext(bak)[1][1:]) + 1
+        bak = stem + f'.{num}'
+    copy_file(file, bak, keepmeta=keepmeta)
+
+
+def recover_file(file, bakdir=None, suffix=None, keepmeta=True):
+    """
+    recover file from numeric backup in bakdir or same dir
+    """
+    bak_dir = bakdir if bakdir else osp.dirname(file)
+    assert osp.isdir(bak_dir)
+    bn = osp.basename(file)
+    files = glob.glob(osp.join(bak_dir, f'{bn}.*'))
+    if not files:
+        raise FileNotFoundError(f'No backup found for {file} under {bak_dir}')
+    if suffix:
+        bak = osp.join(bak_dir, bn+suffix)
+        copy_file(bak, file, keepmeta=keepmeta)
+        return
+    latest = max([int(num_sfx) for file in files if (num_sfx := osp.splitext(file)[1][1:]).isnumeric()])
+    bak = osp.join(bak_dir, f'{bn}.{latest}')
+    copy_file(bak, file, keepmeta=keepmeta)
+
+
+def deprecate_log(replacewith=None):
+    replacement = replacewith if replacewith else 'a documented replacement'
+    return f'This is deprecated; use {replacement} instead'
+
+
+def load_lines(path, rmlineend=False):
+    with open(path) as fp:
+        lines = fp.readlines()
+        if rmlineend:
+            lines = [line.rstrip('\n').rstrip('\r') for line in lines]
+    return lines
+
+
+def save_lines(path, lines, toappend=False, addlineend=True, style='posix'):
+    if isinstance(lines, str):
+        lines = [lines]
+    lines_to_write = copy.deepcopy(lines)
+    mode = 'a' if toappend else 'w'
+    if addlineend:
+        line_end = '\n' if style == 'posix' else '\r\n'
+        lines_to_write = [line+line_end for line in lines]
+    with open(path, mode) as fp:
+        fp.writelines(lines_to_write)
+    return lines_to_write
+
+
+def load_text(path):
+    with open(path) as fp:
+        text = fp.read()
+    return text
+
+
+def save_text(path, text, toappend=False):
+    assert isinstance(text, str)
+    mode = 'a' if toappend else 'w'
+    with open(path, mode) as fp:
+        fp.write(text)
+
+
 def _test():
     pass
 
